#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <stdint.h>
#include <linux/dvb/dmx.h>
#include "filter.h"


#define MAX_SECTION_SIZE 8192
#define MAX_PIDS 1000

#if !defined(DMX_FILTER_SIZE)
#define DMX_FILTER_SIZE 16
#endif





void parse_pmt(int fd)
{
	uint8_t buf[MAX_SECTION_SIZE];
	int bytes;

	unsigned char stream_type;
	unsigned char table_id, section_number, last_section_number;
	unsigned int section_length, pcr_pid, prog_desc = 0;

	unsigned int section_syntax_indicator, program_number, version_number, reserved;
	unsigned int es_descriptor, found_prog_ca_desc = 0, found_stream_ca_desc = 0;
	unsigned int program_info_length, current_next_indicator;
	unsigned int es_pid, es_info_length = 0;
	unsigned int temp = 0, i = 0;
	
	bytes = read(fd, buf, sizeof(buf));
	if (bytes < 0) {
		perror("read");
		if (errno != EOVERFLOW)
			exit(1);
	}


	
	table_id = (unsigned int) buf[0];
	section_syntax_indicator = buf[1] & 0x80;
	reserved = buf[1] & 0x0c;
	section_length = 0;
	section_length = ((section_length | (buf[1] & 0x0f)) << 8) | buf[2];
	program_number = 0;
	program_number = ((program_number | buf[3]) << 8) | buf[4];
	reserved = buf[5] & 0xc0;
	version_number = buf[5] & 0x3e;
	current_next_indicator = buf[5] & 0x01;
	section_number = (unsigned int) buf[6];
	last_section_number = (unsigned int) buf[7];
	reserved = buf[8] & 0xe0;
	pcr_pid = 0;
	pcr_pid = ((pcr_pid | (buf[8] & 0x1f)) << 8) | buf[9];
	reserved = buf[10] & 0xf0;
	program_info_length = 0;
	program_info_length = ((program_info_length | (buf[10] & 0x0f)) << 8) | buf[11]; 


	printf ("Table ID=[%d]\n", table_id);
	printf ("Section Syntax=[%d]\n", section_syntax_indicator);
	printf ("Section Length=[%d]\n", section_length);
	printf ("Program Number=[%d]\n", program_number);
	printf ("Version Number=[%d]\n", version_number);
	printf ("Current Next=[%d]\n", current_next_indicator);
	printf ("Section Number=[%d]\n", section_number);
	printf ("Last Section=[%d]\n", last_section_number);
	printf ("PCR PID=[%d]\n", pcr_pid);
	printf ("Program Length=[%d]\n", program_info_length);


	if (program_info_length) {

	    printf ("\n\nProgram Descriptors=[");
	    for (i = 0; i < program_info_length; i++) {
		printf ("%d ", buf[12 + i]);
		if (buf[12 + i] == 0x09) {
		    found_prog_ca_desc = 1;
		}
		prog_desc = i + 1;
	    }
	    
	    printf ("]\n\n");
	    if (found_prog_ca_desc) {
	    	printf ("Found CA descriptor @ program level\n");
	    }

	}
	
//	i = 12;
	i = prog_desc + 12;
	
	while ( i < (section_length - 4) ) {
	
		stream_type = (unsigned int) buf[i];
		reserved = buf[i + 1] & 0xe0;
		es_pid = 0;
		es_pid = ((es_pid | (buf[i + 1] & 0x1f)) << 8) | buf[i + 2];
		reserved = (buf[i + 3] & 0xf0);
		es_info_length = 0;
		es_info_length = (es_info_length | (buf[i + 3] & 0x0f) << 8) | buf[i + 4]; 

		i = i + 5;
		
		printf ("\ti=[%d] ES Type=[%d] ES pid=[%d] ES length=[%d]\n", i, stream_type, es_pid, es_info_length);
		
		if (es_info_length != 0) {

			printf ("\n\nES Descriptors = [");
			
			while (i < (temp + es_info_length + 17)) {				
				es_descriptor = buf[i];	

				if (es_descriptor == 0x09) {
					found_stream_ca_desc = 1;
				}
				
				printf (" %02x", es_descriptor);
				i++;
				
				if ( i == (section_length - 1) ) {
					break;
				}
			}
			temp = temp + es_info_length + 5;			
			printf ("]\n\n");


			if (found_stream_ca_desc) {
				printf ("Found CA desccriptor @ Stream level\n");					
			}
		}
	}	
}


void usage(void)
{
	fprintf(stderr, "usage: pmt_parse PMT_PID\n");
	exit(1);
}



int main(int argc, char *argv[])
{
	int dmxfd;
	unsigned long pid = 0, tid;
	char * dmxdev = "/dev/dvb/adapter0/demux0";
        unsigned char filter[DMX_FILTER_SIZE];
        unsigned char mask[DMX_FILTER_SIZE];

	int i, entries = 0;
	
	if (argc < 2 || argc > 3)
		usage();

	i = 0, entries = 0;
	/*	Read line after line till end		*/
	pid = strtoul (argv[1], NULL, 0);

	if (pid > 0x1fff)
		usage();
	tid = 0x100;

	memset(filter, '\0', sizeof(filter));
	memset(mask, '\0', sizeof(mask));

	if ((dmxfd = open(dmxdev, O_RDWR)) < 0){
		perror("open");
		return 1;
	}

	if (set_filter(dmxfd, pid, filter, mask) != 0)
		return 1;

	parse_pmt(dmxfd);
	close(dmxfd);

	return 0;
}
